=== modified file '.bzrignore'
--- .bzrignore	2008-02-18 21:17:42 +0000
+++ .bzrignore	2008-07-17 03:46:13 +0000
@@ -40,6 +40,7 @@
 doc/developers/performance.png
 bzrlib/_dirstate_helpers_c.c
 bzrlib/_knit_load_data_c.c
+bzrlib/_walkdirs_win32.c
 doc/en/release-notes/NEWS.txt
 doc/en/developer-guide/HACKING.txt
 doc/en/user-reference/bzr_man.txt

=== modified file 'NEWS'
--- NEWS	2008-07-17 09:58:00 +0000
+++ NEWS	2008-07-17 12:58:30 +0000
@@ -10,6 +10,39 @@
 
   FEATURES:
 
+
+  IMPROVEMENTS:
+
+    * Implemented a custom ``walkdirs_utf8`` implementation for win32.
+      This uses a pyrex extension to get direct access to the
+      ``FindFirstFileW`` style apis, rather than using ``listdir`` +
+      ``lstat``. Shows a very strong improvement in commands like
+      ``status`` and ``diff`` which have to iterate the working tree.
+      Anywhere from 2x-6x faster depending on the size of the tree (bigger
+      trees, bigger benefit.) (John Arbash Meinel)
+
+
+  BUG FIXES:
+
+
+  DOCUMENTATION:
+
+
+  TESTING:
+
+
+  API CHANGES:
+
+
+  INTERNALS:
+
+
+
+bzr 1.6beta3 2008-07-17
+-----------------------
+
+  FEATURES:
+
     * New ``pre_change_branch_tip`` hook that is called before the
       branch tip is moved, while the branch is write-locked.  See the User
       Reference for signature details.  (Andrew Bennetts)

=== added file 'bzrlib/_walkdirs_win32.h'
--- bzrlib/_walkdirs_win32.h	1970-01-01 00:00:00 +0000
+++ bzrlib/_walkdirs_win32.h	2008-07-17 03:46:13 +0000
@@ -0,0 +1,25 @@
+/*
+ * Copyright (C) 2007 Canonical Ltd
+ * 
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
+
+/* Header includes, etc for _walkdirs_win32
+ * Pyrex doesn't support #define, and defining WIN32_LEAN_AND_MEAN makes
+ * importing windows.h a lot less painful.
+ */
+
+#define WIN32_LEAN_AND_MEAN
+#include <windows.h>

=== added file 'bzrlib/_walkdirs_win32.pyx'
--- bzrlib/_walkdirs_win32.pyx	1970-01-01 00:00:00 +0000
+++ bzrlib/_walkdirs_win32.pyx	2008-07-17 12:21:19 +0000
@@ -0,0 +1,283 @@
+# Copyright (C) 2008 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Helper functions for Walkdirs on win32."""
+
+
+cdef extern from "_walkdirs_win32.h":
+    cdef struct _HANDLE:
+        pass
+    ctypedef _HANDLE *HANDLE
+    ctypedef unsigned long DWORD
+    ctypedef long long __int64
+    ctypedef unsigned short WCHAR
+    cdef struct _FILETIME:
+        DWORD dwHighDateTime
+        DWORD dwLowDateTime
+    ctypedef _FILETIME FILETIME
+
+    cdef struct _WIN32_FIND_DATAW:
+        DWORD dwFileAttributes
+        FILETIME ftCreationTime
+        FILETIME ftLastAccessTime
+        FILETIME ftLastWriteTime
+        DWORD nFileSizeHigh
+        DWORD nFileSizeLow
+        # Some reserved stuff here
+        WCHAR cFileName[260] # MAX_PATH
+        WCHAR cAlternateFilename[14]
+
+    # We have to use the typedef trick, otherwise pyrex uses:
+    #  struct WIN32_FIND_DATAW
+    # which fails due to 'incomplete type'
+    ctypedef _WIN32_FIND_DATAW WIN32_FIND_DATAW
+
+    cdef HANDLE INVALID_HANDLE_VALUE
+    cdef HANDLE FindFirstFileW(WCHAR *path, WIN32_FIND_DATAW *data)
+    cdef int FindNextFileW(HANDLE search, WIN32_FIND_DATAW *data)
+    cdef int FindClose(HANDLE search)
+
+    cdef DWORD FILE_ATTRIBUTE_READONLY
+    cdef DWORD FILE_ATTRIBUTE_DIRECTORY
+    cdef int ERROR_NO_MORE_FILES
+
+    cdef int GetLastError()
+
+    # Wide character functions
+    DWORD wcslen(WCHAR *)
+
+
+cdef extern from "Python.h":
+    WCHAR *PyUnicode_AS_UNICODE(object)
+    Py_ssize_t PyUnicode_GET_SIZE(object)
+    object PyUnicode_FromUnicode(WCHAR *, Py_ssize_t)
+    int PyList_Append(object, object) except -1
+    object PyUnicode_AsUTF8String(object)
+
+
+import operator
+import stat
+
+from bzrlib import osutils
+
+
+cdef class _Win32Stat:
+    """Represent a 'stat' result generated from WIN32_FIND_DATA"""
+
+    cdef readonly int st_mode
+    cdef readonly double st_ctime
+    cdef readonly double st_mtime
+    cdef readonly double st_atime
+    cdef readonly __int64 st_size
+
+    # os.stat always returns 0, so we hard code it here
+    cdef readonly int st_dev
+    cdef readonly int st_ino
+
+    def __repr__(self):
+        """Repr is the same as a Stat object.
+
+        (mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime)
+        """
+        return repr((self.st_mode, 0, 0, 0, 0, 0, self.st_size, self.st_atime,
+                     self.st_mtime, self.st_ctime))
+
+
+cdef object _get_name(WIN32_FIND_DATAW *data):
+    """Extract the Unicode name for this file/dir."""
+    return PyUnicode_FromUnicode(data.cFileName,
+                                 wcslen(data.cFileName))
+
+
+cdef int _get_mode_bits(WIN32_FIND_DATAW *data):
+    cdef int mode_bits
+
+    mode_bits = 0100666 # writeable file, the most common
+    if data.dwFileAttributes & FILE_ATTRIBUTE_READONLY == FILE_ATTRIBUTE_READONLY:
+        mode_bits ^= 0222 # remove the write bits
+    if data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY == FILE_ATTRIBUTE_DIRECTORY:
+        # Remove the FILE bit, set the DIR bit, and set the EXEC bits
+        mode_bits ^= 0140111
+    return mode_bits
+
+
+cdef __int64 _get_size(WIN32_FIND_DATAW *data):
+    # Pyrex casts a DWORD into a PyLong anyway, so it is safe to do << 32
+    # on a DWORD
+    return ((<__int64>data.nFileSizeHigh) << 32) + data.nFileSizeLow
+
+
+cdef double _ftime_to_timestamp(FILETIME *ft):
+    """Convert from a FILETIME struct into a floating point timestamp.
+
+    The fields of a FILETIME structure are the hi and lo part
+    of a 64-bit value expressed in 100 nanosecond units.
+    1e7 is one second in such units; 1e-7 the inverse.
+    429.4967296 is 2**32 / 1e7 or 2**32 * 1e-7.
+    It also uses the epoch 1601-01-01 rather than 1970-01-01
+    (taken from posixmodule.c)
+    """
+    cdef __int64 val
+    # NB: This gives slightly different results versus casting to a 64-bit
+    #     integer and doing integer math before casting into a floating
+    #     point number. But the difference is in the sub millisecond range,
+    #     which doesn't seem critical here.
+    # secs between epochs: 11,644,473,600
+    val = ((<__int64>ft.dwHighDateTime) << 32) + ft.dwLowDateTime
+    return (val * 1.0e-7) - 11644473600.0
+
+
+cdef int _should_skip(WIN32_FIND_DATAW *data):
+    """Is this '.' or '..' so we should skip it?"""
+    if (data.cFileName[0] != c'.'):
+        return 0
+    if data.cFileName[1] == c'\0':
+        return 1
+    if data.cFileName[1] == c'.' and data.cFileName[2] == c'\0':
+        return 1
+    return 0
+
+
+cdef class Win32Finder:
+    """A class which encapsulates the search of files in a given directory"""
+
+    cdef object _top
+    cdef object _prefix
+
+    cdef object _directory_kind
+    cdef object _file_kind
+
+    cdef object _pending
+    cdef object _last_dirblock
+
+    def __init__(self, top, prefix=""):
+        self._top = top
+        self._prefix = prefix
+
+        self._directory_kind = osutils._directory_kind
+        self._file_kind = osutils._formats[stat.S_IFREG]
+
+        self._pending = [(osutils.safe_utf8(prefix), osutils.safe_unicode(top))]
+        self._last_dirblock = None
+
+    def __iter__(self):
+        return self
+
+    cdef object _get_kind(self, WIN32_FIND_DATAW *data):
+        if data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY:
+            return self._directory_kind
+        return self._file_kind
+
+    cdef _Win32Stat _get_stat_value(self, WIN32_FIND_DATAW *data):
+        """Get the filename and the stat information."""
+        cdef _Win32Stat statvalue
+
+        statvalue = _Win32Stat()
+        statvalue.st_mode = _get_mode_bits(data)
+        statvalue.st_ctime = _ftime_to_timestamp(&data.ftCreationTime)
+        statvalue.st_mtime = _ftime_to_timestamp(&data.ftLastWriteTime)
+        statvalue.st_atime = _ftime_to_timestamp(&data.ftLastAccessTime)
+        statvalue.st_size = _get_size(data)
+        statvalue.st_ino = 0
+        statvalue.st_dev = 0
+        return statvalue
+
+    def _get_files_in(self, directory, relprefix):
+        cdef WIN32_FIND_DATAW search_data
+        cdef HANDLE hFindFile
+        cdef int last_err
+        cdef WCHAR *query
+        cdef int result
+
+        top_star = directory + '*'
+
+        dirblock = []
+
+        query = PyUnicode_AS_UNICODE(top_star)
+        hFindFile = FindFirstFileW(query, &search_data)
+        if hFindFile == INVALID_HANDLE_VALUE:
+            # Raise an exception? This path doesn't seem to exist
+            raise WindowsError(GetLastError(), top_star)
+
+        try:
+            result = 1
+            while result:
+                # Skip '.' and '..'
+                if _should_skip(&search_data):
+                    result = FindNextFileW(hFindFile, &search_data)
+                    continue
+                name_unicode = _get_name(&search_data)
+                name_utf8 = PyUnicode_AsUTF8String(name_unicode)
+                relpath = relprefix + name_utf8
+                abspath = directory + name_unicode
+                PyList_Append(dirblock, 
+                    (relpath, name_utf8, 
+                     self._get_kind(&search_data),
+                     self._get_stat_value(&search_data),
+                     abspath))
+
+                result = FindNextFileW(hFindFile, &search_data)
+            # FindNextFileW sets GetLastError() == ERROR_NO_MORE_FILES when it
+            # actually finishes. If we have anything else, then we have a
+            # genuine problem
+            last_err = GetLastError()
+            if last_err != ERROR_NO_MORE_FILES:
+                raise WindowsError(last_err)
+        finally:
+            result = FindClose(hFindFile)
+            if result == 0:
+                last_err = GetLastError()
+                pass
+        return dirblock
+
+    cdef _update_pending(self):
+        """If we had a result before, add the subdirs to pending."""
+        if self._last_dirblock is not None:
+            # push the entries left in the dirblock onto the pending queue
+            # we do this here, because we allow the user to modified the
+            # queue before the next iteration
+            for d in reversed(self._last_dirblock):
+                if d[2] == self._directory_kind:
+                    self._pending.append((d[0], d[-1]))
+            self._last_dirblock = None
+        
+    def __next__(self):
+        self._update_pending()
+        if not self._pending:
+            raise StopIteration()
+        relroot, top = self._pending.pop()
+        # NB: At the moment Pyrex doesn't support Unicode literals, which means
+        # that all of these string literals are going to be upcasted to Unicode
+        # at runtime... :(
+        # Maybe we could use unicode(x) during __init__?
+        if relroot:
+            relprefix = relroot + '/'
+        else:
+            relprefix = ''
+        top_slash = top + '/'
+
+        dirblock = self._get_files_in(top_slash, relprefix)
+        dirblock.sort(key=operator.itemgetter(1))
+        self._last_dirblock = dirblock
+        return (relroot, top), dirblock
+
+
+def _walkdirs_utf8_win32_find_file(top, prefix=""):
+    """Implement a version of walkdirs_utf8 for win32.
+
+    This uses the find files api to both list the files and to stat them.
+    """
+    return Win32Finder(top, prefix=prefix)

=== modified file 'bzrlib/osutils.py'
--- bzrlib/osutils.py	2008-06-26 17:18:55 +0000
+++ bzrlib/osutils.py	2008-07-17 12:55:11 +0000
@@ -53,6 +53,7 @@
     )
 """)
 
+
 import bzrlib
 from bzrlib import symbol_versioning
 from bzrlib.symbol_versioning import (
@@ -1193,8 +1194,14 @@
         pass to os functions to affect the file in question. (such as os.lstat)
     """
     fs_encoding = _fs_enc.upper()
-    if (sys.platform == 'win32' or
-        fs_encoding not in ('UTF-8', 'US-ASCII', 'ANSI_X3.4-1968')): # ascii
+    if sys.platform == 'win32':
+        try:
+            from bzrlib._walkdirs_win32 import _walkdirs_utf8_win32_find_file
+        except ImportError:
+            return _walkdirs_unicode_to_utf8(top, prefix=prefix)
+        else:
+            return _walkdirs_utf8_win32_find_file(top, prefix=prefix)
+    if fs_encoding not in ('UTF-8', 'US-ASCII', 'ANSI_X3.4-1968'): # ascii
         return _walkdirs_unicode_to_utf8(top, prefix=prefix)
     else:
         return _walkdirs_fs_utf8(top, prefix=prefix)

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2008-07-14 07:47:45 +0000
+++ bzrlib/tests/__init__.py	2008-07-17 12:55:11 +0000
@@ -2796,6 +2796,7 @@
                    'bzrlib.tests.test_versionedfile',
                    'bzrlib.tests.test_version',
                    'bzrlib.tests.test_version_info',
+                   'bzrlib.tests.test__walkdirs_win32',
                    'bzrlib.tests.test_weave',
                    'bzrlib.tests.test_whitebox',
                    'bzrlib.tests.test_win32utils',

=== added file 'bzrlib/tests/test__walkdirs_win32.py'
--- bzrlib/tests/test__walkdirs_win32.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test__walkdirs_win32.py	2008-07-17 03:46:13 +0000
@@ -0,0 +1,122 @@
+# Copyright (C) 2008 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for the win32 walkdir extension."""
+
+import errno
+
+from bzrlib import tests
+
+
+class _WalkdirsWin32Feature(tests.Feature):
+
+    def _probe(self):
+        try:
+            import bzrlib._walkdirs_win32
+        except ImportError:
+            return False
+        else:
+            return True
+
+    def feature_name(self):
+        return 'bzrlib._walkdirs_win32'
+
+WalkdirsWin32Feature = _WalkdirsWin32Feature()
+
+
+class TestWin32Finder(tests.TestCaseInTempDir):
+
+    _test_needs_features = [WalkdirsWin32Feature]
+
+    def setUp(self):
+        super(TestWin32Finder, self).setUp()
+        from bzrlib._walkdirs_win32 import (
+            _walkdirs_utf8_win32_find_file
+            )
+        self.walkdirs_utf8 = _walkdirs_utf8_win32_find_file
+
+    def _remove_stat_from_dirblock(self, dirblock):
+        return [info[:3] + info[4:] for info in dirblock]
+
+    def assertWalkdirs(self, expected, top, prefix=''):
+        finder = self.walkdirs_utf8(top, prefix=prefix)
+        result = []
+        for dirname, dirblock in finder:
+            result.append((dirname, self._remove_stat_from_dirblock(dirblock)))
+        self.assertEqual(expected, result)
+
+    def test_empty_directory(self):
+        self.assertWalkdirs([(('', u'.'), [])], u'.')
+
+    def test_file_in_dir(self):
+        self.build_tree(['foo'])
+        self.assertWalkdirs([
+            (('', u'.'), [('foo', 'foo', 'file', u'./foo')])
+            ], u'.')
+
+    def test_subdir(self):
+        self.build_tree(['foo', 'bar/', 'bar/baz'])
+        self.assertWalkdirs([
+            (('', u'.'), [('bar', 'bar', 'directory', u'./bar'),
+                          ('foo', 'foo', 'file', u'./foo'),
+                         ]),
+            (('bar', u'./bar'), [('bar/baz', 'baz', 'file', u'./bar/baz')]),
+            ], '.')
+        self.assertWalkdirs([
+            (('xxx', u'.'), [('xxx/bar', 'bar', 'directory', u'./bar'),
+                             ('xxx/foo', 'foo', 'file', u'./foo'),
+                            ]),
+            (('xxx/bar', u'./bar'), [('xxx/bar/baz', 'baz', 'file', u'./bar/baz')]),
+            ], '.', prefix='xxx')
+        self.assertWalkdirs([
+            (('', u'bar'), [('baz', 'baz', 'file', u'bar/baz')]),
+            ], 'bar')
+
+    def test_skip_subdir(self): 
+        self.build_tree(['a/', 'b/', 'c/', 'a/aa', 'b/bb', 'c/cc'])
+        base_dirblock = [('a', 'a', 'directory', u'./a'),
+                          ('b', 'b', 'directory', u'./b'),
+                          ('c', 'c', 'directory', u'./c'),
+                         ]
+        self.assertWalkdirs([
+            (('', u'.'), base_dirblock),
+            (('a', u'./a'), [('a/aa', 'aa', 'file', u'./a/aa')]),
+            (('b', u'./b'), [('b/bb', 'bb', 'file', u'./b/bb')]),
+            (('c', u'./c'), [('c/cc', 'cc', 'file', u'./c/cc')]),
+            ], '.')
+
+        walker = self.walkdirs_utf8('.')
+        dir_info, first_dirblock = walker.next()
+        self.assertEqual(('', u'.'), dir_info)
+        self.assertEqual(base_dirblock,
+                         self._remove_stat_from_dirblock(first_dirblock))
+        # Now, remove 'b' and it should be skipped on the next round
+        del first_dirblock[1]
+        dir_info, second_dirblock = walker.next()
+        second_dirblock = self._remove_stat_from_dirblock(second_dirblock)
+        self.assertEqual(('a', u'./a'), dir_info)
+        self.assertEqual([('a/aa', 'aa', 'file', u'./a/aa')], second_dirblock)
+        dir_info, third_dirblock = walker.next()
+        third_dirblock = self._remove_stat_from_dirblock(third_dirblock)
+        self.assertEqual(('c', u'./c'), dir_info)
+        self.assertEqual([('c/cc', 'cc', 'file', u'./c/cc')], third_dirblock)
+
+    def test_missing_dir(self):
+        e = self.assertRaises(WindowsError, list,
+                                self.walkdirs_utf8(u'no_such_dir'))
+        self.assertEqual(errno.ENOENT, e.errno)
+        self.assertEqual(3, e.winerror)
+        self.assertEqual((3, u'no_such_dir/*'), e.args)

=== modified file 'bzrlib/tests/test_osutils.py'
--- bzrlib/tests/test_osutils.py	2008-06-11 03:56:46 +0000
+++ bzrlib/tests/test_osutils.py	2008-07-17 12:21:19 +0000
@@ -16,16 +16,19 @@
 
 """Tests for the osutils wrapper."""
 
+from cStringIO import StringIO
 import errno
 import os
 import socket
 import stat
 import sys
+import time
 
 import bzrlib
 from bzrlib import (
     errors,
     osutils,
+    tests,
     win32utils,
     )
 from bzrlib.errors import BzrBadParameterNotUnicode, InvalidURL
@@ -46,7 +49,8 @@
 from bzrlib.tests.file_utils import (
     FakeReadFile,
     )
-from cStringIO import StringIO
+from bzrlib.tests.test__walkdirs_win32 import WalkdirsWin32Feature
+
 
 class TestOSUtils(TestCaseInTempDir):
 
@@ -974,6 +978,99 @@
         self._filter_out_stat(result)
         self.assertEqual(expected_dirblocks, result)
 
+    def test__walkdirs_utf_win32_find_file(self):
+        self.requireFeature(WalkdirsWin32Feature)
+        self.requireFeature(tests.UnicodeFilenameFeature)
+        from bzrlib._walkdirs_win32 import _walkdirs_utf8_win32_find_file
+        name0u = u'0file-\xb6'
+        name1u = u'1dir-\u062c\u0648'
+        name2u = u'2file-\u0633'
+        tree = [
+            name0u,
+            name1u + '/',
+            name1u + '/' + name0u,
+            name1u + '/' + name1u + '/',
+            name2u,
+            ]
+        self.build_tree(tree)
+        name0 = name0u.encode('utf8')
+        name1 = name1u.encode('utf8')
+        name2 = name2u.encode('utf8')
+
+        # All of the abspaths should be in unicode, all of the relative paths
+        # should be in utf8
+        expected_dirblocks = [
+                (('', '.'),
+                 [(name0, name0, 'file', './' + name0u),
+                  (name1, name1, 'directory', './' + name1u),
+                  (name2, name2, 'file', './' + name2u),
+                 ]
+                ),
+                ((name1, './' + name1u),
+                 [(name1 + '/' + name0, name0, 'file', './' + name1u
+                                                        + '/' + name0u),
+                  (name1 + '/' + name1, name1, 'directory', './' + name1u
+                                                            + '/' + name1u),
+                 ]
+                ),
+                ((name1 + '/' + name1, './' + name1u + '/' + name1u),
+                 [
+                 ]
+                ),
+            ]
+        result = list(_walkdirs_utf8_win32_find_file(u'.'))
+        self._filter_out_stat(result)
+        self.assertEqual(expected_dirblocks, result)
+
+    def assertStatIsCorrect(self, path, win32stat):
+        os_stat = os.stat(path)
+        self.assertEqual(os_stat.st_size, win32stat.st_size)
+        self.assertAlmostEqual(os_stat.st_mtime, win32stat.st_mtime, places=4)
+        self.assertAlmostEqual(os_stat.st_ctime, win32stat.st_ctime, places=4)
+        self.assertAlmostEqual(os_stat.st_atime, win32stat.st_atime, places=4)
+        self.assertEqual(os_stat.st_dev, win32stat.st_dev)
+        self.assertEqual(os_stat.st_ino, win32stat.st_ino)
+        self.assertEqual(os_stat.st_mode, win32stat.st_mode)
+
+    def test__walkdirs_utf_win32_find_file_stat_file(self):
+        """make sure our Stat values are valid"""
+        self.requireFeature(WalkdirsWin32Feature)
+        self.requireFeature(tests.UnicodeFilenameFeature)
+        from bzrlib._walkdirs_win32 import _walkdirs_utf8_win32_find_file
+        name0u = u'0file-\xb6'
+        name0 = name0u.encode('utf8')
+        self.build_tree([name0u])
+        # I hate to sleep() here, but I'm trying to make the ctime different
+        # from the mtime
+        time.sleep(2)
+        f = open(name0u, 'ab')
+        try:
+            f.write('just a small update')
+        finally:
+            f.close()
+
+        result = list(_walkdirs_utf8_win32_find_file(u'.'))
+        entry = result[0][1][0]
+        self.assertEqual((name0, name0, 'file'), entry[:3])
+        self.assertEqual(u'./' + name0u, entry[4])
+        self.assertStatIsCorrect(entry[4], entry[3])
+        self.assertNotEqual(entry[3].st_mtime, entry[3].st_ctime)
+
+    def test__walkdirs_utf_win32_find_file_stat_directory(self):
+        """make sure our Stat values are valid"""
+        self.requireFeature(WalkdirsWin32Feature)
+        self.requireFeature(tests.UnicodeFilenameFeature)
+        from bzrlib._walkdirs_win32 import _walkdirs_utf8_win32_find_file
+        name0u = u'0dir-\u062c\u0648'
+        name0 = name0u.encode('utf8')
+        self.build_tree([name0u + '/'])
+
+        result = list(_walkdirs_utf8_win32_find_file(u'.'))
+        entry = result[0][1][0]
+        self.assertEqual((name0, name0, 'directory'), entry[:3])
+        self.assertEqual(u'./' + name0u, entry[4])
+        self.assertStatIsCorrect(entry[4], entry[3])
+
     def assertPathCompare(self, path_less, path_greater):
         """check that path_less and path_greater compare correctly."""
         self.assertEqual(0, osutils.compare_paths_prefix_order(

=== modified file 'setup.py'
--- setup.py	2008-07-02 16:52:12 +0000
+++ setup.py	2008-07-17 12:55:11 +0000
@@ -226,6 +226,8 @@
 
 add_pyrex_extension('bzrlib._dirstate_helpers_c')
 add_pyrex_extension('bzrlib._knit_load_data_c')
+if sys.platform == 'win32':
+    add_pyrex_extension('bzrlib._walkdirs_win32')
 ext_modules.append(Extension('bzrlib._patiencediff_c', ['bzrlib/_patiencediff_c.c']))
 
 

